#include <stdio.h>
#include <string.h>
#include <math.h>
#include "clFFT.h"

#define max(a,b) (a < b) ? b : a
#define min(a,b) (a > b) ? b : a

/////////////////////////////////////
// OpenCL FFT 3D function ///////////
/////////////////////////////////////

int FFT_3D_OpenCL(float *tab_real, float *tab_img, const char* direction, int size)
{
  int i;

  /* OpenCL variables. */
  cl_int err;
  cl_platform_id platform = 0;
  cl_device_id device = 0;
  cl_context context = 0;
  cl_command_queue queue = 0;

  /* Input and Output  buffer. */
  cl_mem buffersIn[2]  = {0, 0};
  cl_mem buffersOut[2] = {0, 0};

  /* Temporary buffer. */
  cl_mem tmpBuffer = 0;

  /* Size of temp buffer. */
  size_t tmpBufferSize = 0;
  int status = 0;
  int ret = 0;

  /* Size of vectors. */
  size_t N = size;
  
  /* Total input length */
  int length = N*N*N;

  /* FFT library realted declarations. */
  clfftPlanHandle planHandle;
  clfftDim dim = CLFFT_3D;
  size_t clLengths[3] = {N, N, N};

  /* Setup OpenCL environment. */
  err = clGetPlatformIDs(1, &platform, NULL);
  err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);

  /* Create an OpenCL context. */
  context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);

  /* Create a command queue. */
  queue = clCreateCommandQueueWithProperties(context, device, 0, &err);

  /* Setup clFFT. */
  clfftSetupData fftSetup;
  err = clfftInitSetupData(&fftSetup);
  err = clfftSetup(&fftSetup);

  /* Create a default plan for a complex FFT. */
  err = clfftCreateDefaultPlan(&planHandle, context, dim, clLengths);

  /* Set plan parameters. */
  err = clfftSetPlanPrecision(planHandle, CLFFT_SINGLE);
  err = clfftSetLayout(planHandle, CLFFT_COMPLEX_PLANAR, CLFFT_COMPLEX_PLANAR);
  err = clfftSetResultLocation(planHandle, CLFFT_OUTOFPLACE);

  /* Bake the plan. */
  err = clfftBakePlan(planHandle, 1, &queue, NULL, NULL);

  /* Real and Imaginary arrays. */
  cl_float* inReal  = (cl_float*) malloc (length * sizeof (cl_float));
  cl_float* inImag  = (cl_float*) malloc (length * sizeof (cl_float));
  cl_float* outReal = (cl_float*) malloc (length * sizeof (cl_float));
  cl_float* outImag = (cl_float*) malloc (length * sizeof (cl_float));

  /* Initialization of inReal, inImag, outReal and outImag. */
  for (i = 0; i < length; i++)
  {
    inReal[i] = tab_real[i];
    inImag[i] = tab_img[i];
    outReal[i] = 0.0f;
    outImag[i] = 0.0f;
  }  

  /* Create temporary buffer. */
  status = clfftGetTmpBufSize(planHandle, &tmpBufferSize);

  if ((status == 0) && (tmpBufferSize > 0)) {
    tmpBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, tmpBufferSize, 0, &err);
    if (err != CL_SUCCESS)
      printf("Error with tmpBuffer clCreateBuffer\n");
  }

  /* Prepare OpenCL memory objects : create buffer for input. */
  buffersIn[0] = clCreateBuffer( context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length * sizeof(cl_float), inReal, &err);
  if (err != CL_SUCCESS)
    printf("Error with buffersIn[0] clCreateBuffer into forward FFT\n");

  /* Enqueue write tab array into buffersIn[0]. */
  err = clEnqueueWriteBuffer(queue, buffersIn[0], CL_TRUE, 0, length * sizeof(float), inReal, 0, NULL, NULL);
  if (err != CL_SUCCESS)
    printf("Error with buffersIn[0] clEnqueueWriteBuffer into forward FFT\n");

  /* Prepare OpenCL memory objects : create buffer for input. */
  buffersIn[1] = clCreateBuffer( context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length * sizeof(cl_float), inImag, &err);
  if (err != CL_SUCCESS)
    printf("Error with buffersIn[1] clCreateBuffer into forward FFT\n");

  /* Enqueue write tab array into buffersIn[1]. */
  err = clEnqueueWriteBuffer(queue, buffersIn[1], CL_TRUE, 0, length *
      sizeof(float),
      inImag, 0, NULL, NULL);
  if (err != CL_SUCCESS)
    printf("Error with buffersIn[1] clEnqueueWriteBuffer into forward FFT\n");

  /* Prepare OpenCL memory objects : create buffer for output. */
  buffersOut[0] = clCreateBuffer (context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length * sizeof(cl_float), outReal, &err);
  if (err != CL_SUCCESS)
    printf("Error with buffersOut[0] clCreateBuffer\n");

  buffersOut[1] = clCreateBuffer (context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, length * sizeof(cl_float), outImag, &err);
  if (err != CL_SUCCESS)
    printf("Error with buffersOut[1] clCreateBuffer\n");

  /* Execute Forward or Backward FFT. */
  if(strcmp(direction,"forward") == 0)
  {
    /* Execute the plan. */
    err = clfftEnqueueTransform(planHandle, CLFFT_FORWARD, 1, &queue, 0, NULL, NULL,
	buffersIn, buffersOut, tmpBuffer);
    printf("fft forward\n");
    if (err != CL_SUCCESS)
      printf("Error with clfftEnqueueTransform Forward\n");
  }	
  /* Enqueue transform for backward FFT. */	
  else if(strcmp(direction,"backward") == 0)
  {
    /* Execute the plan. */
    err = clfftEnqueueTransform(planHandle, CLFFT_BACKWARD, 1, &queue, 0, NULL, NULL,
	buffersIn, buffersOut, tmpBuffer);
    printf("fft backward\n");
    if (err != CL_SUCCESS)
      printf("Error with clfftEnqueueTransform Backward\n");
  }

  /* Wait for calculations to be finished. */
  err = clFinish(queue);
  if (err != CL_SUCCESS)
    printf("Error with clFinish\n");

  /* Fetch results of calculation: Real and Imaginary. */
  err = clEnqueueReadBuffer( queue, buffersOut[0], CL_TRUE, 0, length * sizeof(float), tab_real, 0, NULL, NULL);
  if (err != CL_SUCCESS)
    printf("Error with clEnqueueReadBuffer\n");
  err = clEnqueueReadBuffer( queue, buffersOut[1], CL_TRUE, 0, length * sizeof(float), tab_img, 0, NULL, NULL);
  if (err != CL_SUCCESS)
    printf("Error with clEnqueueReadBuffer\n");

  /* Release OpenCL memory objects. */
  clReleaseMemObject(buffersIn[0]);
  clReleaseMemObject(buffersIn[1]);
  clReleaseMemObject(buffersOut[0]);
  clReleaseMemObject(buffersOut[1]);
  clReleaseMemObject(tmpBuffer);

  /* Release the plan. */
  err = clfftDestroyPlan(&planHandle);
  if (err != CL_SUCCESS)
    printf("Error with clfftDestroyPlan\n");

  /* Release clFFT library. */
  clfftTeardown();

  /* Release OpenCL working objects. */
  clReleaseCommandQueue(queue);
  clReleaseContext(context);

  return ret;
}

int main() {

  // File for input positions 
  FILE *fileInput;

  // File for output potential 
  FILE *fileOutput;

  // Indexes 
  int i, j, k, l, node_x, node_y, node_z;

  // Arrays for positions
  float *x, *y, *z;

  // Array for final solution
  float *potential_real, *potential_img;
  float ***potential;
  float ***potential_shifted;

  // Arrays for Green function, density and potential
  float *green_real, *green_img; 
  float *density_real, *density_img;
  float ***green;
  float ***density;

  // For CIC computation
  double space_x, space_y, space_z, dx, dy, dz, tx, ty, tz, frac_x, frac_y, frac_z;

  // Interpolation method : set "NGP" or "CIC"
  const char* interpolation = "CIC";

  // Number of stars in galaxy
  double numStars = 10240;

  // Size of the grid side
  int Ng = 256; 

  // Gravitational constant
  // G = 6.67428E-8 cm^3 g^-1 s^-2
  double G = 6.67428E-8;            

  // Factor for FFT, useful to stay in float range
  // taking 10^10 solar masses
  double factor_fft = 1.989E43;  

  // 1kpc = 3.085678 10^(19) m = 3.085678 10^(21) cm = 3.085678 10^(16) km
  double kpc = 3.085678E21;   

  // mass of each star (with numStars = 10240) : 10^7 * solar mass = 10^7 * 2.0 10^(33) g
  double mass = 1.0E7 * 2.0E33;

  // For adaptative size of cell into grid
  double Xmin, Ymin, Zmin, Xmax, Ymax, Zmax;

  // Neighboor for calculating charge assignment function
  int neighboor_x, neighboor_y, neighboor_z;

  // Correction for computing cell space into grid
  int add_correction_x, add_correction_y, add_correction_z;

  // Epsilon factors for calculating charge assignment function
  int epsilon_x, epsilon_y, epsilon_z;

  // Allocating all arrays 

  // Allocating x, y, z positions of stars
  if (!(x = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for x positions.\n");
    return -1;
  }
  if (!(y = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for y positions.\n");
    return -1;
  }
  if (!(z = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for z positions.\n");
    return -1;
  }

  // Allocating final solution array
  // x axis
  if (!(potential = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for final potential.\n");
    return -1;
  }
  for (i = 0; i < Ng; ++i) {
    // y-axis
    if (!(potential[i] = calloc(Ng,sizeof(float *)))) {
      fprintf(stderr,"Unable to create potential y axis.\n");
      return -1;
    }
    // z-axis
    for (j = 0; j < Ng; ++j) {
      if (!(potential[i][j] = calloc(Ng, sizeof(float)))) {
	fprintf(stderr,"Unable to create potential z axis.\n");
	return -1;
      }
    }
  }

  // Allocating real and imaginary parts for potential
  if (!(potential_real = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for potential_real buffer.\n");
    return -1;
  }
  if (!(potential_img = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for potential_img buffer.\n");
    return -1;
  }

  // Allocating density array
  // x axis
  if (!(density = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for density.\n");
    return -1;
  }
  for (i = 0; i < Ng; ++i) {
    // y-axis
    if (!(density[i] = calloc(Ng,sizeof(float *)))) {
      fprintf(stderr,"Unable to create density y axis.\n");
      return -1;
    }
    // z-axis
    for (j = 0; j < Ng; ++j) {
      if (!(density[i][j] = calloc(Ng, sizeof(float)))) {
	fprintf(stderr,"Unable to create density z axis.\n");
	return -1;
      }
    }
  }

  // Allocating real and imaginary parts for density
  if (!(density_real = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for potential_real buffer.\n");
    return -1;
  }
  if (!(density_img = calloc(Ng*Ng*Ng, sizeof(float)))) {
    fprintf(stderr,"Unable to allocate space for potential_img buffer.\n");
    return -1;
  }

  // Allocating Green function array on grid
  // x axis
  if (!(green = calloc(Ng, sizeof(float *)))) {
    fprintf(stderr,"Unable to create Green's function x axis.\n");
    return -1;
  }
  for (i = 0; i < Ng; ++i) {
    // y-axis
    if (!(green[i] = calloc(Ng,sizeof(float *)))) {
      fprintf(stderr,"Unable to create Green's function y axis.\n");
      return -1;
    }
    // z-axis
    for (j = 0; j < Ng; ++j) {
      if (!(green[i][j] = calloc(Ng, sizeof(float)))) {
	fprintf(stderr,"Unable to create Green's function z axis.\n");
	return -1;
      }
    }
  }

  // Allocating real and imaginary parts for green function
  if (!(green_real = calloc(pow(Ng,3), sizeof(float *)))) {
    fprintf(stderr,"Unable to create Green's function x axis.\n");
    return -1;
  }
  if (!(green_img = calloc(pow(Ng,3), sizeof(float *)))) {
    fprintf(stderr,"Unable to create Green's function x axis.\n");
    return -1;
  }

  // Allocating temporary array for shift backward FFT
  // x axis
  if (!(potential_shifted=calloc(Ng,sizeof(float *)))) {
    fprintf(stderr,"Unable to create temp potential x axis.\n");
    return -1;
  }
  for (i = 0; i < Ng; ++i) {
    // y-axis
    if (!(potential_shifted[i] = calloc(Ng,sizeof(float *)))) {
      fprintf(stderr,"Unable to create temp potential y axis.\n");
      return -1;
    }
    // z-axis
    for (j = 0; j < Ng; ++j) {
      if (!(potential_shifted[i][j] = calloc(Ng,sizeof(float)))) {
	fprintf(stderr,"Unable to create temp potential z axis.\n");
	return -1;
      }
    }
  }

  // Check for bad grids
  if (Ng <= 0) {
    fprintf(stderr,"Grid dimensions must be greater than zero!");
    return -1;
  }

  // Reading the positions of stars from Positions_Disk.dat
  fileInput = fopen("Positions_Disk.dat","r");
  if ( fileInput != NULL )
  {
    for (i=0; i<numStars; i++)
      fscanf(fileInput,"%f %f %f\n", &x[i], &y[i], &z[i]);
  } 
  fclose(fileInput);

  // Find min and max of galaxy with stars positions:
  Xmin = 0.0; Ymin = 0.0; Zmin = 0.0;
  Xmax = 0.0; Ymax = 0.0; Zmax = 0.0;

  for (i = 0; i < numStars; i++) 
  {
    Xmax = max(x[i], Xmax); 
    Xmin = min(x[i], Xmin); 
    Ymax = max(y[i], Ymax); 
    Ymin = min(y[i], Ymin); 
    Zmax = max(z[i], Zmax); 
    Zmin = min(z[i], Zmin); 
  }

  // Compute higher absolute value
  // and define corrections 
  if (Xmax > fabs(Xmin))
  { space_x = Xmax/(Ng/2);
    add_correction_x = -1;
  }
  else { space_x = fabs(Xmin)/(Ng/2);
    add_correction_x = 0;
  }

  if (Ymax > fabs(Ymin))
  { space_y = Ymax/(Ng/2);
    add_correction_y = -1;
  }
  else { space_y = fabs(Ymin)/(Ng/2);
    add_correction_y = 0;
  }

  if (Zmax > fabs(Zmin))
  { space_z = Zmax/(Ng/2);
    add_correction_z = -1;
  }
  else { space_z = fabs(Zmin)/(Ng/2);
    add_correction_z = 0;
  }

  // Print size of cell : space_x, space_y, space_z
  fprintf(stderr,"Grid cell spacings (kpc): x = %f y = %f z = %f\n",
      space_x,space_y,space_z);

  // Compute coordinates on the 3D mesh
  for (i = 0; i < numStars; i++) {
    x[i] = x[i]/(space_x) + (double) (Ng/2 + add_correction_x);
    y[i] = y[i]/(space_y) + (double) (Ng/2 + add_correction_y);
    z[i] = z[i]/(space_z) + (double) (Ng/2 + add_correction_z);

  }

  // Compute the density using the CIC routine. The positions are shifted 

  // Initialization - Set density to zero 
  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) 
	   density[i][j][k] = 0.0f;

  // 2 available methods : "Nearest Grid point" (NGP) or "Cloud In Cell" (CIC)
  if (strcmp(interpolation,"NGP")==0) {
    for (i = 0; i < numStars; i++) {

       // Figure out which node owns the particle
       node_x = (int) x[i];
       node_y = (int) y[i];
       node_z = (int) z[i];
       
       // Cumulate density 
       density[node_x][node_y][node_z] += mass/(factor_fft*(space_x*space_y*space_z));
    }
  }
  else if (strcmp(interpolation,"CIC")==0)
         { 
	 for (i = 0; i < numStars; i++) 
            { 
	    // Figure out which node owns the particle
	    node_x = (int) x[i];
	    node_y = (int) y[i];
	    node_z = (int) z[i];
 
	    // Set the CIC size fractions
	    dx = fmod(abs(x[i]),space_x)/space_x;
	    dy = fmod(abs(y[i]),space_y)/space_y;
	    dz = fmod(abs(z[i]),space_z)/space_z;
	    tx = 1.0 - dx;
	    ty = 1.0 - dy;
	    tz = 1.0 - dz;
 
	    // Cases for CIC fractions
	    // Version with (i,j) centered on cells
	    if ((dx < 0.5) && (node_x == 0))
	      {frac_x = dx;
	       neighboor_x = 0;
	       epsilon_x = 0;}
	    else if ((dx > 0.5) && (node_x == Ng-1))
		   {frac_x = dx - 0.5;
		    neighboor_x = 0;
		    epsilon_x = 0;}
	    else if (dx < 0.5) 
		   {frac_x = dx + 0.5;
		    neighboor_x = -1;
		    epsilon_x = 1;}
	    else {frac_x = tx + 0.5;
	        neighboor_x = 1;
	        epsilon_x = 1;}
 
	    if ((dy < 0.5) && (node_y == 0))
	      {frac_y = dy;
	       neighboor_y = 0;
	       epsilon_y = 0;}
	    else if ((dy > 0.5) && (node_y == Ng-1))
		   {frac_y = dy - 0.5;
		    neighboor_y = 0;
		    epsilon_y = 0;}
	    else if (dy < 0.5) 
		   {frac_y = dy + 0.5;
		    neighboor_y = -1;
		    epsilon_y = 1;}
	   else {frac_y = ty + 0.5;
	       neighboor_y = 1;
	       epsilon_y = 1;}
 
	    if ((dz < 0.5) && (node_z == 0))
	      {frac_z = dz;
	       neighboor_z = 0;
	       epsilon_z = 0;}
	   else if ((dz > 0.5) && (node_z == Ng-1))
		  {frac_z = dz - 0.5;
		   neighboor_z = 0;
		   epsilon_z = 0;}
	   else if (dz < 0.5) 
		  {frac_z = dz + 0.5;
		   neighboor_z = -1;
		   epsilon_z = 1;}
	   else {frac_z = tz + 0.5;
	       neighboor_z = 1;
	       epsilon_z = 1;}
 
	   // Calculate the CIC densities
	   density[node_x][node_y][node_z] += mass*(frac_x*frac_y*frac_z)/((space_x*space_y*space_z)*factor_fft);
	   density[node_x+neighboor_x][node_y][node_z] += mass*(epsilon_x*(1.0-frac_x)*frac_y*frac_z)/((space_x*space_y*space_z)*factor_fft);
	   density[node_x][node_y+neighboor_y][node_z] += mass*(frac_x*epsilon_y*(1.0-frac_y)*frac_z)/((space_x*space_y*space_z)*factor_fft);
	   density[node_x][node_y][node_z+neighboor_z] += mass*(frac_x*frac_y*epsilon_z*(1.0-frac_z))/((space_x*space_y*space_z)*factor_fft);
	   density[node_x][node_y+neighboor_y][node_z+neighboor_z] += mass*(frac_x*epsilon_y*(1.0-frac_y)*epsilon_z*(1.0-frac_z))/((space_x*space_y*space_z)*factor_fft);
	   density[node_x+neighboor_x][node_y+neighboor_y][node_z] += mass*(epsilon_x*(1.0-frac_x)*epsilon_y*(1.0-frac_y)*frac_z)/((space_x*space_y*space_z)*factor_fft);
	   density[node_x+neighboor_x][node_y][node_z+neighboor_z] += mass*(epsilon_x*(1.0-frac_x)*frac_y*epsilon_z*(1.0-frac_z))/((space_x*space_y*space_z)*factor_fft);
	   density[node_x+neighboor_x][node_y+neighboor_y][node_z+neighboor_z] += mass*(epsilon_x*(1.0-frac_x)*epsilon_y*(1.0-frac_y)*epsilon_z*(1.0-frac_z))/((space_x*space_y*space_z)*factor_fft);
 
   }
  }
  // Define Green's function
  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) 
           { 
            dx = sqrt((i-Ng/2-add_correction_x)*(i-Ng/2-add_correction_x))*space_x;
            dy = sqrt((j-Ng/2-add_correction_y)*(j-Ng/2-add_correction_y))*space_y;
            dz = sqrt((k-Ng/2-add_correction_z)*(k-Ng/2-add_correction_z))*space_z;

            // Discretized Green's function
            green[i][j][k] = -1.0/(4.0*M_PI*sqrt(dx*dx + dy*dy + dz*dz));
           }    

  // Set singularity in Green's function to 1.0
  green[Ng/2+add_correction_x][Ng/2+add_correction_y][Ng/2+add_correction_z] = 1.0;

  // Fprintf Green's function 
  FILE* greenFile;
  greenFile = fopen("Green_function.dat","w");
  for (i = 0; i < Ng; i++)
     for (j = 0; j < Ng; j++)
        for (k = 0; k < Ng; k++)
           fprintf(greenFile,"%d %d %d %e\n", i, j, k, green[i][j][k]);
  fclose(greenFile);

  // Fprintf Density function 
  FILE* densityFile;
  densityFile = fopen("Density_function.dat","w");
  for (i = 0; i < Ng; i++)
     for (j = 0; j < Ng; j++)
        for (k = 0; k < Ng; k++)
           fprintf(densityFile,"%d %d %d %e\n", i, j, k, density[i][j][k]);
  fclose(densityFile);

  // Pack Green's function and the density into 1D arrays
  l = 0;
  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) {
           // For Green 3D
           green_real[l] = green[i][j][k];
           green_img[l] = 0.0f;

           // For density 3D
           density_real[l] = density[i][j][k];
           density_img[l] = 0.0f;
           l++;
           }

  // Execute Forward FFT for Density grid and Green function
  FFT_3D_OpenCL(density_real, density_img, "forward", Ng);
  FFT_3D_OpenCL(green_real, green_img, "forward", Ng);

  // Multiply the density by Green's function to find the k-space density and 
  // invert for the real density. 
  for (i = 0; i < Ng*Ng*Ng; i++) {
     potential_real[i] = green_real[i]*density_real[i]-green_img[i]*density_img[i];
     potential_img[i] = green_real[i]*density_img[i]+green_img[i]*density_real[i];
  }

  // Backward FFT for product for Poisson equation into Fourier space
  FFT_3D_OpenCL(potential_real, potential_img, "backward", Ng);

  // Multiply FFT(potential) by 4*pi*G
  l = 0;
  for (i = 0; i < Ng; i++) {
     for (j = 0; j < Ng; j++) {
        for (k = 0; k < Ng; k++) {
           potential[i][j][k] = 4.0*M_PI*G*potential_real[l]*factor_fft*space_x*space_y*space_z/kpc;
           l++;
        }
     }
  }

  // Shift the potential like fftshift Matlab
  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) 
           {
            // Shift 3D potential like with fftshift Matlab

	    // 7 to 1
	    if (i< Ng/2 && j < Ng/2 && k < Ng/2)
	      potential_shifted[i][j][k]=potential[i+Ng/2][j+Ng/2][k+Ng/2];
	    // 1 to 7	
	    if (i >= Ng/2 && j >= Ng/2 && k >= Ng/2)	
	      potential_shifted[i][j][k]=potential[i-Ng/2][j-Ng/2][k-Ng/2];

	    // 8 to 2
	    if (i >= Ng/2 && j < Ng/2 && k < Ng/2)	
	      potential_shifted[i][j][k]=potential[i-Ng/2][j+Ng/2][k+Ng/2];
	    // 2 to 8
	    if (i < Ng/2 && j >= Ng/2 && k >= Ng/2)	
	      potential_shifted[i][j][k]=potential[i+Ng/2][j-Ng/2][k-Ng/2];

	    // 5 to 3
	    if (i >= Ng/2 && j >= Ng/2 && k < Ng/2)	
	      potential_shifted[i][j][k]=potential[i-Ng/2][j-Ng/2][k+Ng/2];
	    // 3 to 5	
	    if (i < Ng/2 && j < Ng/2 && k >= Ng/2)	
	      potential_shifted[i][j][k]=potential[i+Ng/2][j+Ng/2][k-Ng/2];

	    // 6 to 4
	    if (i < Ng/2 && j >= Ng/2 && k < Ng/2)	
	      potential_shifted[i][j][k]=potential[i+Ng/2][j-Ng/2][k+Ng/2];
	    // 4 to 6
	    if (i >= Ng/2 && j < Ng/2 && k >= Ng/2)	
	      potential_shifted[i][j][k]=potential[i-Ng/2][j+Ng/2][k-Ng/2];
	   }

  // Copy shifted potential into potential
  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) 
           potential[i][j][k]=potential_shifted[i][j][k];


  // Save potential into file
  fileOutput=fopen("Potential_3D.dat","w");
  fprintf(fileOutput,"%d %d %d %d\n", add_correction_x, add_correction_y, add_correction_z, 0);

  for (i = 0; i < Ng; i++) 
     for (j = 0; j < Ng; j++) 
        for (k = 0; k < Ng; k++) 
           fprintf(fileOutput,"%f %f %f %3.6e\n",(double) i, (double) j, (double) k, potential[i][j][k]);

  fclose(fileOutput);

  for (i = 0; i < Ng; ++i) {
     for (j = 0; j < Ng; ++j) {
        free(green[i][j]);
        free(density[i][j]);
        free(potential[i][j]);
        free(potential_shifted[i][j]);
     }
  free(green[i]); 
  free(density[i]);
  free(potential[i]);
  free(potential_shifted[i]); 
  }

  free(green);
  free(density);
  free(potential);
  free(potential_shifted);

  free(green_real);
  free(green_img);
  free(density_real);
  free(density_img);
  free(potential_real);
  free(potential_img);

  return 0;
}
